#include "RealmBackgroundScroller.h"
#include <SDL2/SDL.h>
#include "const7.h"
#include "RealmSceneBackground.h"
#include "RealmMapDisplay.h"
#include "RealmSceneView.h"

RealmBackgroundScroller::RealmBackgroundScroller(int hex_radius, int grid_size, bool inverted, int maps_radius):
	CurrentCenterCoordinate(0)
{
	InnerSpiralSize = 1;
	GridSize = grid_size;
	Inverted = inverted;
	HexRadius = hex_radius;
	MapsRadius = maps_radius;

	updateDimensions();



	NeedsUpdate = true;
	TextureSizeChanged = true;

	// based on given radius, calcuate the number of hexes in the map
	int n;
	for (n = 1; n <= MapsRadius; n++) {
		InnerSpiralSize += 6 * n;
	}
	OuterSpiralSize = InnerSpiralSize + 6 * n;

	Texture = new SDL_Texture* [OuterSpiralSize];

	for (n = 0; n < OuterSpiralSize; n++) {
		Texture[n] = nullptr;
	}

	TextureScreenLocation = new XYLocation[InnerSpiralSize];
}

RealmBackgroundScroller::~RealmBackgroundScroller() {
	destroyTextures();
	delete Texture;
}

void RealmBackgroundScroller::setCenter(HexSpiralCoordinate new_center, bool force_update) {
	if (CurrentCenterCoordinate.getHexIndex() != new_center.getHexIndex() || force_update) {
		NeedsUpdate = true;
		CurrentCenterCoordinate = new_center;
	}
}

int RealmBackgroundScroller::getTextureIndex(int i) {
	HexSpiralCoordinate wrapped_center = CurrentCenterCoordinate.findWrappedPosition(MapsRadius + 1, !Inverted);
	HexSpiralCoordinate result_abs_coord = wrapped_center + HexSpiralCoordinate(i);

	return result_abs_coord.findWrappedPosition(MapsRadius + 1, !Inverted).getHexIndex();
}

void RealmBackgroundScroller::setHexRadius(int hex_radius) {
	if (HexRadius != hex_radius) {
		HexRadius = hex_radius;
		TextureSizeChanged = true;
		NeedsUpdate = true;
	}
}

void RealmBackgroundScroller::makeTextures(SDL_Renderer* renderer) {
	for (int n = 0; n < OuterSpiralSize; n++) {
		Texture[n] = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET, MapWidth, MapHeight);
	}
}

void RealmBackgroundScroller::destroyTextures() {
	for (int n = 0; n < OuterSpiralSize; n++) {
		if (Texture[n])
			SDL_DestroyTexture(Texture[n]);
	}
}

void RealmBackgroundScroller::updateDimensions() {
	// set width and height
	if (!Inverted) {
		MapHeight = (GridSize + 1) * 2 * SQRT_3 * HexRadius;
		MapWidth = (GridSize + 1) * 2 * 1.5 * HexRadius;
	} else {
		MapWidth = (GridSize + 1) * 2 * SQRT_3 * HexRadius;
		MapHeight = (GridSize + 1) * 2 * 1.5 * HexRadius;
	}
}

void RealmBackgroundScroller::populateTextures(SDL_Renderer* renderer) {
	RealmSceneBackground* scenebg;
	int texture_index;

	for (int i = 0; i < InnerSpiralSize; i++) {
		long int mapidx = (HexSpiralCoordinate(i) + CurrentCenterCoordinate).getHexIndex();

		scenebg = MapDisplay->getSceneBackground(mapidx);
		texture_index = getTextureIndex(i);


		if (scenebg)
			scenebg->drawToTexture(renderer, Texture[texture_index], NeedsUpdate);
		else {
			SDL_SetRenderTarget(renderer, Texture[texture_index]);
			SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x00, 0x00);
			SDL_RenderClear(renderer);
			SDL_SetTextureBlendMode(Texture[texture_index], SDL_BLENDMODE_BLEND);
			SDL_SetRenderTarget(renderer, NULL);
		}

	}

	NeedsUpdate = false;
}


void RealmBackgroundScroller::update(RealmSceneView* scene_view) {
	RealmSceneBackground* scenebg;
	setCenter(scene_view->getHexLocation().
				getParentSpiralCoordinate(!Inverted));

	for (int i = 0; i < InnerSpiralSize; i++) {

		long int mapidx = (HexSpiralCoordinate(i) + CurrentCenterCoordinate).getHexIndex();

		scenebg = MapDisplay->getSceneBackground(mapidx);

		if (scenebg) {
			scenebg->setSize(scene_view->getBaseRadius());
			if (scenebg->needsUpdate())
				NeedsUpdate = true;
		}

		XYLocation MapCoord = HexSpiralCoordinate(mapidx)
				.getChildSpiralCoordinate(Inverted)
				.getXY(Inverted);

		TextureScreenLocation[i] = scene_view->getScreenLocation(MapCoord);
	}
}

void RealmBackgroundScroller::updateTextures(SDL_Renderer* renderer) {

	if (TextureSizeChanged) {
		// destroy and recreate textures
		destroyTextures();
		updateDimensions();
		makeTextures(renderer);
		TextureSizeChanged = false;
	}

	if (NeedsUpdate)
		populateTextures(renderer);
}

void RealmBackgroundScroller::draw(SDL_Renderer* renderer) {
	SDL_Rect dstrect;
	int texture_index;

	for (int i = 0; i < InnerSpiralSize; i++) {
		double zoom_factor = MapDisplay->getZoomFactor();
		int zoomed_map_width = MapWidth * zoom_factor;
		int zoomed_map_height = MapHeight * zoom_factor;

		dstrect.x = TextureScreenLocation[i].getx() - zoomed_map_width / 2;
		dstrect.y = TextureScreenLocation[i].gety() - zoomed_map_height / 2;
		dstrect.w = zoomed_map_width;
		dstrect.h = zoomed_map_height;

		texture_index = getTextureIndex(i);

		SDL_RenderCopy(renderer, Texture[texture_index], NULL, &dstrect);
	}
}
