

#include "RealmView.h"
#include "HexGridDisplay.h"
#include "const7.h"

#include "RealmCharacterGroup.h"
#include "RealmCharacterGroupPositioner.h"

#include <iostream>

#include <math.h>
#include <stdlib.h>

#include <SDL2/SDL2_gfxPrimitives.h>

#include <cstdlib>

using namespace std;

RealmView::RealmView(XYLocation center, int window_width, int window_height):
			BaseRadius(48),
			TerrainController(),
			SceneView(1, 5, 1, false, TerrainController.getMapRepository(), this),
			ZoneView(2, 5, 1,  true, TerrainController.getMapRepository(), this),
			RegionView(3, 4, 2, false, TerrainController.getMapRepository(), this),
			RealmWorldView(4, 1, 1, true, TerrainController.getMapRepository(), this),
			CharacterGroup(this)
 {
	WindowCenter = center;
	AnimatingMove = AnimatingZoom = false;

	ViewMode = 4;

	XYLocation windowsize(window_width, window_height);

	// CoordinateMap.setBaseRadius(BaseRadius);
	// CoordinateMap.setCenter(WindowCenter);
	// CoordinateMap.setScreenSize(windowsize);

	SceneView.setSuperGridOffset(3);

	ZoneView.setSubGridOffset(3);
	ZoneView.setSuperGridOffset(4);
	ZoneView.setSuperZoomOffset(4);

	RegionView.setSubZoomOffset(4);
	RegionView.setSubGridOffset(3);
	RegionView.setSuperGridOffset(3);
	RegionView.setSuperZoomOffset(2);

	RealmWorldView.setSubGridOffset(3);
	RealmWorldView.setSubZoomOffset(2);
}


void RealmView::init() {

	TerrainController.init(BaseRadius);

	updateScene(1);
	updateScene(2);
	updateScene(3);
	updateScene(4);
}

void RealmView::cleanup() {
	TerrainController.cleanup();
}

bool RealmView::isInvertedZoom() {
	return RealmLocation::isInvertedGrid(ViewMode);
}

RealmSceneView* RealmView::getFocusedRealmSceneView() {
	switch (ViewMode) {
		case 1: return &SceneView; break;
		case 2: return &ZoneView; break;
		case 3: return &RegionView; break;
		case 4: return &RealmWorldView; break;
	}
	return nullptr; // shouldn't ever get here
}

RealmSceneView& RealmView::getRealmSceneView(int view_level)
{
	return (view_level == 1 ? SceneView :
			view_level == 2 ? ZoneView :
				view_level == 3 ? RegionView :
					RealmWorldView);
}

void RealmView::cycleSelectedObject() {

	getFocusedRealmSceneView()->cycleSelectedObject();
}

void RealmView::activateMapDisplay() {
	if (AnimatingMove)
		return;

	getFocusedRealmSceneView()->toggleMapDisplay();
}

void RealmView::mapPopulateStep() {
	if (AnimatingMove)
		return;

	getFocusedRealmSceneView()->mapPopulateStep();
}


void RealmView::increaseBaseRadius() {
	BaseRadius++;
	SceneView.updateBaseRadius();
	ZoneView.updateBaseRadius();
	RegionView.updateBaseRadius();
	RealmWorldView.updateBaseRadius();
}

void RealmView::decreaseBaseRadius() {
	if (BaseRadius > 1) {
		BaseRadius--;
		SceneView.updateBaseRadius();
		ZoneView.updateBaseRadius();
		RegionView.updateBaseRadius();
		RealmWorldView.updateBaseRadius();
	}
}



// beginMovementKey to be called to start a movement action when a key is pressed and held
void RealmView::beginMovementKey(int key_direction) {
	getFocusedRealmSceneView()->beginMovementKey(key_direction);
}

// stopMovement is called when a key is released
void RealmView::stopMovementKey(int key_direction) {
	getFocusedRealmSceneView()->stopMovementKey(key_direction);
}

void RealmView::moveFinished(int source_view_mode, HexSpiralCoordinate targetpos) {
	AnimatingMove = false;


	//XYLocation worldcoord = ViewCenter.getRealmWorldCoordinate();
	//cout << "New Realm World Location (global coords): " << worldcoord.getx() << "," << worldcoord.gety() << endl;

	ViewCenter.setHexIndex(targetpos.getHexIndex(), source_view_mode);

	if (source_view_mode > 1) {
		ViewCenter.setHexIndex(0, source_view_mode-1);
		if (source_view_mode > 2)
			getRealmSceneView(source_view_mode-1).setCameraPosition( ViewCenter.getHexSpiralCoordinate(source_view_mode-1));
		updateScene(source_view_mode - 1);
	}

	if (source_view_mode == 1) {
		ZoneView.setSubGridPosition(ViewCenter.getXY(1));
	}
	if (source_view_mode == 2) {
		RegionView.setSubGridPosition(ViewCenter.getXY(2));
	}
	if (source_view_mode == 3) {
		RealmWorldView.setSubGridPosition(ViewCenter.getXY(3));
	}
}

void RealmView::updateScene(int view_level)
{
	RealmSceneView* curview =
							view_level == 1 ? &SceneView :
							view_level == 2 ? &ZoneView :
							view_level == 3 ? &RegionView :
											&RealmWorldView;

	curview->reCenterCurrentMap();

	TerrainController.updateScene(view_level, ViewCenter,
									curview
								);
}


// todo move into more suitable object than TerrainController
void RealmView::autoUpdateObjects() {
	// refactoring of experimental features coming soon....
	// TerrainController.autoMove();
}


void RealmView::zoomIn() {
	// initiates a zoom in, to be shown and completed during the update loop

	// zoom in process:
	// quick-fade in the hexgrid / objects on the map at the next zoom level,
	// with radius and scale sized down from the base radius by the zoom factor (typically 7^(-3) hex ratio formula)

	// Then begin the zoom by incrementing the inner radius during each loop update
	// until the inner radius reaches the size of the base radius
	// fade out the (now large) grid at the oversize level

	// right now best not to allow simultaneous zoom and movement, as it could cause odd/undefined behavior
	if (AnimatingMove)
		return;

	if (ViewMode == 1) {
		// already zoomed in as close as possible
	} else if (ViewMode == 2) {
		TargetViewMode = 1;
		ZoneView.disableMapDisplay();
		SceneView.reCenterCurrentMap();
		SceneView.activateMapDisplay();

		ZoneView.setSubGridPosition(ViewCenter.getXY(1));

		SceneView.startZoom(2);

		// begin character group zone-to-scene zoom-in transition phase 1
		CharacterGroup.getPositioner()->startTransition(INTO_SCENE_PHASE1);

	} else if (ViewMode == 3) {
		TargetViewMode = 2;
		ZoneView.reCenterCurrentMap();
		ZoneView.activateMapDisplay();

		RegionView.setSubGridPosition(ViewCenter.getXY(2));

		ZoneView.startZoom(1);
		RegionView.startZoom(-2);

		CharacterGroup.getPositioner()->startTransition(INTO_ZONE);
	} else if (ViewMode == 4) {
		RegionView.reCenterCurrentMap();
		RegionView.activateMapDisplay();
		TargetViewMode = 3;

		RealmWorldView.setSubGridPosition(ViewCenter.getXY(3));

		// ZoneView.startZoom(5);
		RegionView.startZoom(2);
		RealmWorldView.startZoom(-1);
		CharacterGroup.getPositioner()->startTransition(INTO_REGION);
	}

	AnimatingZoom = true;
}


void RealmView::zoomOut() {

	// initiates a zoom out, to be shown and completed during the update loop

	// right now best not to allow simultaneous zoom and movement, as it could cause odd/undefined behavior
	if (AnimatingMove)
		return;

	if (ViewMode == 1) {
		TargetViewMode = 2;
		ViewCenter.setScene(
			(
				ViewCenter.getHexSpiralCoordinate(2) +
				ViewCenter.getHexSpiralCoordinate(1)
					.getParentSpiralCoordinate()
			).getHexIndex()
		);
		ZoneView.setCameraPosition(ViewCenter.getHexSpiralCoordinate(2));
		ZoneView.setSubGridPosition(
			(
				ViewCenter.getHexSpiralCoordinate(1) -
				ViewCenter.getHexSpiralCoordinate(1)
					.getParentSpiralCoordinate()
					.getChildSpiralCoordinate()
			).getXY( false )
		);

		RegionView.setSubGridPosition(
			(
				ViewCenter.getHexSpiralCoordinate(2) -
				ViewCenter.getHexSpiralCoordinate(2)
					.getParentSpiralCoordinate(false)
					.getChildSpiralCoordinate(true)
			).getXY(true)
		);

		CharacterGroup.getPositioner()->setZonePosition(ViewCenter.getHexSpiralCoordinate(2));

		ZoneView.startZoom(0);

	} else if (ViewMode == 2) {
		TargetViewMode = 3;
		ViewCenter.setZone(
			(
				ViewCenter.getHexSpiralCoordinate(3) +
				ViewCenter.getHexSpiralCoordinate(2)
					.getParentSpiralCoordinate(false)
			).getHexIndex()
		);
		RegionView.setCameraPosition(ViewCenter.getHexSpiralCoordinate(3));
		RegionView.setSubGridPosition(
			(
				ViewCenter.getHexSpiralCoordinate(2) -
				ViewCenter.getHexSpiralCoordinate(2)
					.getParentSpiralCoordinate(false)
					.getChildSpiralCoordinate(true)
			).getXY(true)
		);

		RealmWorldView.setSubGridPosition(
			(
				ViewCenter.getHexSpiralCoordinate(3) -
				ViewCenter.getHexSpiralCoordinate(3)
					.getParentSpiralCoordinate()
					.getChildSpiralCoordinate()
			).getXY(false)
		);

		CharacterGroup.getPositioner()->setRegionPosition(ViewCenter.getHexSpiralCoordinate(3));

		ZoneView.startZoom(5);
		RegionView.startZoom(2);

		CharacterGroup.getPositioner()->startTransition(OUT_ZONE);
	} else if (ViewMode == 3) {
		TargetViewMode = 4;

		ViewCenter.setRegion(
			(
				ViewCenter.getHexSpiralCoordinate(4) +
				ViewCenter.getHexSpiralCoordinate(3)
					.getParentSpiralCoordinate()
			).getHexIndex()
		);

		ViewCenter.setZone(
			(
				ViewCenter.getHexSpiralCoordinate(3)
								-
				ViewCenter.getHexSpiralCoordinate(3)
					.getParentSpiralCoordinate()
					.getChildSpiralCoordinate()
			).getHexIndex()
		);

		RealmWorldView.setCameraPosition(ViewCenter.getHexSpiralCoordinate(4));
		RealmWorldView.setSubGridPosition(
			(
				ViewCenter.getHexSpiralCoordinate(3) -
				ViewCenter.getHexSpiralCoordinate(3)
					.getParentSpiralCoordinate()
					.getChildSpiralCoordinate()
			).getXY(false)
		);

		RegionView.startZoom(4);
		RealmWorldView.startZoom(1);

		CharacterGroup.getPositioner()->startTransition(OUT_REGION);
	}

	AnimatingZoom = true;
}


void RealmView::zoomFinished(int source_view_mode) {

	AnimatingZoom = false;
	ViewMode = TargetViewMode;

	if (source_view_mode == 1) {
		if (TargetViewMode == 1) {
			if (abs(1 - SceneView.getCurrentZoom()) > 0.001) {
				SceneView.startZoom(1, 5);
			}
			ZoneView.startZoom(0, 5);
			AnimatingZoom = true;

			// begin character group zone-to-scene zoom-in transition phase 2
			CharacterGroup.getPositioner()->startTransition(INTO_SCENE_PHASE2);
		}

		if (TargetViewMode == 2) {
			if (abs(3 - SceneView.getCurrentZoom()) < 0.001) {
				SceneView.startZoom(4, 5);
				ZoneView.startZoom(1, 5);
				AnimatingZoom = true;
				CharacterGroup.getPositioner()->startTransition(OUT_SCENE_PHASE2);
			}
		}
	}
	if (source_view_mode == 2) {
		if (TargetViewMode == 2) {
			if (abs(ZoneView.getCurrentZoom()) < 0.001) {
				SceneView.startZoom(3);
				AnimatingZoom = true;
				CharacterGroup.getPositioner()->startTransition(OUT_SCENE_PHASE1);

			} else if (abs(1 - ZoneView.getCurrentZoom()) < 0.001 && SceneView.getCurrentZoom() > 3) {
				//ZoneView.activateMapDisplay();
				CharacterGroup.getPositioner()->startTransition(AT_ZONE);
			}
		}
		if (TargetViewMode == 1) {
			ZoneView.startZoom(-2, 30);
		}
	}

	// all done animating?  make sure the view mode below gets its position reset to center
	if (!AnimatingZoom) {
		getRealmSceneView(TargetViewMode).refreshCameraPosition();

		if (TargetViewMode > 1) {
			if (TargetViewMode < 4) {
				ViewCenter.setHexIndex(0, TargetViewMode - 1);
				getRealmSceneView(TargetViewMode).setSubGridPosition( XYLocation(0, 0) );
			}

			if (TargetViewMode > 2)
				getRealmSceneView(TargetViewMode-1).setCameraPosition( ViewCenter.getHexSpiralCoordinate(TargetViewMode-1));

			CharacterGroup.getPositioner()->startTransition(TargetViewMode == 2 ? AT_ZONE : TargetViewMode == 3 ? AT_REGION : AT_REALM );

			updateScene(TargetViewMode-1);
		} else {
			CharacterGroup.getPositioner()->startTransition(AT_SCENE);
		}
	}
}


void RealmView::update() {

	// called with each tick...
	// if a move is in progress, increment any needed variables before drawing

	// if (AnimatingZoom) {
		// CoordinateMap.setBaseRadius(SceneView.getZoomRadius(1));
	// }

	// updates current position, if animating a zoom or movement
	if (ViewMode == 1 || TargetViewMode == 1 || ViewMode == 2 )
		SceneView.update();
	if (ViewMode == 2 || TargetViewMode == 2 || ViewMode == 3 || ViewMode == 1 || TargetViewMode == 3)
		ZoneView.update();
	if (ViewMode == 3 || TargetViewMode == 3 || ViewMode == 2 || ViewMode == 4)
		RegionView.update();
	if (ViewMode == 4 || TargetViewMode == 4 || ViewMode == 3)
		RealmWorldView.update();

	CharacterGroup.update();

	// will this be needed later?
	// TerrainController.update();
}

void RealmView::updateTextures(SDL_Renderer* renderer) {
	if ( (ViewMode == 1 || TargetViewMode == 1 || ViewMode == 2 ) && TargetViewMode != 3) {
		SceneView.updateTextures(renderer);
	}
	if (ViewMode == 2 || TargetViewMode == 2 || ViewMode == 1  || ViewMode == 3) {
		ZoneView.updateTextures(renderer);
	}
	if (ViewMode == 3 || TargetViewMode == 3 || ViewMode == 2 || ViewMode == 4) {
		RegionView.updateTextures(renderer);
	}
	if (ViewMode == 4 || TargetViewMode == 4 || ViewMode == 3) {
		RealmWorldView.updateTextures(renderer);
	}
}

void RealmView::draw(SDL_Renderer* renderer) {

	/* reset black background */
	SDL_SetRenderDrawColor(renderer, 0, 0, 0, SDL_ALPHA_OPAQUE);
	SDL_RenderClear(renderer);

	if ( (ViewMode == 1 || TargetViewMode == 1 || ViewMode == 2 ) && TargetViewMode != 3) {
		SceneView.draw(renderer);
	}
	if (ViewMode == 2 || TargetViewMode == 2 || ViewMode == 1  || ViewMode == 3) {
		ZoneView.draw(renderer);
	}
	if (ViewMode == 3 || TargetViewMode == 3 || ViewMode == 2 || ViewMode == 4) {
		RegionView.draw(renderer);
	}
	if (ViewMode == 4 || TargetViewMode == 4 || ViewMode == 3) {
		RealmWorldView.draw(renderer);
	}

	CharacterGroup.draw(renderer);

//	if (ViewMode == 1 && !AnimatingZoom)
//		CoordinateMap.draw(renderer);
};


